//
// Verilog parallel Odd-Even sort network
// for W wide data and N inputs.
// For this module N must be a power of two.
//
// Finds max and min of N input signals of W bits wide
// using only combinatorial logic.
//
// Due to the use of arrays in the port section,
// this code needs a System-verilog simulator/compiler
// 

// 
// Freeware 2018, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
//
// Synthesis on Xilinx for Zynq system with W=8, N=8 
// uses 328 LUTS.
//

/*
  The code below builds an Odd-Even sorting network
  instancing a series of two port max/min sorting modules.
  It only outputs the maximum and minum ports, leaving it up to 
  the synthesis tool to remove unused logic. 

     +---+               +---+                 +---+               +---+    
  ---|   |---------------|   |--------------- -|   |---------------|   |---Max
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
     +---+  \  +---+  /  +---+  \  +---+  /    +---+  \  +---+  /  +---+  
             +-|   |-+           +-|   |-+             +-|   |-+          
             +-|   |-+           +-|   |-+             +-|   |-+          
     +---+  /  +---+  \  +---+  /  +---+  \    +---+  /  +---+  \  +---+  
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
     +---+  \  +---+  /  +---+  \  +---+  /    +---+  \  +---+  /  +---+  
             +-|   |-+           +-|   |-+             +-|   |-+          
             +-|   |-+           +-|   |-+             +-|   |-+          
     +---+  /  +---+  \  +---+  /  +---+  \    +---+  /  +---+  \  +---+  
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
     +---+  \  +---+  /  +---+  \  +---+  /    +---+  \  +---+  /  +---+  
             +-|   |-+           +-|   |-+             +-|   |-+          
             +-|   |-+           +-|   |-+             +-|   |-+          
     +---+  /  +---+  \  +---+  /  +---+  \    +---+  /  +---+  \  +---+  
  ---|   |-+           +-|   |-+           +- -|   |-+           +-|   |-X
  ---|   |---------------|   |--------------- -|   |---------------|   |---Min
     +---+               +---+                 +---+               +---+   


 */

module maxmin_N
#(parameter W = 8, // Data width
            N = 8) // Number of entries: power of two
(
   input  [W-1:0] i [0:N-1], // Input signals
   output [W-1:0] max,
   output [W-1:0] min
);   

wire [W-1:0] w [0:N*(N+1)-1];


genvar c,r,a;

generate
   for (c=0; c<N; c=c+1)
   begin
      if (c[0]==1'b0)
      begin // even
         for (r=0; r<N; r=r+2)
            sort2 sort2_i (.a(w[c*N+r]),.b(w[c*N+r+1]),.big(w[c*N+r+N]),.sme(w[c*N+r+N+1]));            
      end
      else
      begin // odd
         assign w[c*N+N]=w[c*N];
         for (r=1; r<N-2; r=r+2)
           sort2 sort2_i (.a(w[c*N+r]),.b(w[c*N+r+1]),.big(w[c*N+r+N]),.sme(w[c*N+r+N+1]));            
         assign w[c*N+N+N-1]=w[c*N+N-1];
      end
   end
   
   // Assign inputs (and no outputs)
   for (r=0; r<N; r=r+1)
   begin
      assign w[r]=i[r];
//      assign o[r]=w[N*N+r];
   end
   
   assign max = w[N*N];
   assign min = w[N*N+N-1];
   
endgenerate   
   


endmodule
